@@ -1,5 +1,4 @@
-# Created: 06.2020
-# Copyright (c) 2020, Matthew Broadway
+# Copyright (c) 2020-2021, Matthew Broadway
 # License: MIT License
 import math
 from typing import Iterable, cast, Union, List, Dict, Callable
@@ -9,15 +8,19 @@ from ezdxf.addons.drawing.properties import (
     RenderContext, VIEWPORT_COLOR, Properties, set_color_alpha, Filling,
 )
 from ezdxf.addons.drawing.text import simplified_text_chunks
-from ezdxf.addons.drawing.utils import get_tri_or_quad_points
+from ezdxf.addons.drawing.type_hints import FilterFunc
 from ezdxf.entities import (
     DXFGraphic, Insert, MText, Polyline, LWPolyline, Spline, Hatch, Attrib,
-    Text, Polyface, Wipeout, AttDef,
+    Text, Polyface, Wipeout, AttDef, Solid, Face3d,
 )
 from ezdxf.entities.dxfentity import DXFTagStorage, DXFEntity
 from ezdxf.layouts import Layout
 from ezdxf.math import Vec3, Z_AXIS
-from ezdxf.render import MeshBuilder, TraceBuilder, Path, make_path, nesting
+from ezdxf.path import (
+    make_path, from_hatch_boundary_path, fast_bbox_detection,
+    winding_deconstruction, from_vertices,
+)
+from ezdxf.render import MeshBuilder, TraceBuilder
 from ezdxf import reorder
 from ezdxf.proxygraphic import ProxyGraphic
 
@@ -42,7 +45,7 @@ class Frontend:
     """
 
     def __init__(self, ctx: RenderContext, out: Backend,
-                 proxy_graphics: int = USE_PROXY_GRAPHICS):
+        proxy_graphics: int = USE_PROXY_GRAPHICS):
         # RenderContext contains all information to resolve resources for a
         # specific DXF document.
         self.ctx = ctx
@@ -76,7 +79,7 @@ class Frontend:
         # sagitta = 1cm
 
         # set to None to disable nested polygon detection:
-        self.nested_polygon_detection = nesting.fast_bbox_detection
+        self.nested_polygon_detection = fast_bbox_detection
 
         self._dispatch = self._build_dispatch_table()
 
@@ -88,10 +91,11 @@ class Frontend:
             'MESH': self.draw_mesh_entity,
             'VIEWPORT': self.draw_viewport_entity,
             'WIPEOUT': self.draw_wipeout_entity,
+            'MTEXT': self.draw_mtext_entity,
         }
         for dxftype in ('LINE', 'XLINE', 'RAY'):
             dispatch_table[dxftype] = self.draw_line_entity
-        for dxftype in ('TEXT', 'MTEXT', 'ATTRIB', 'ATTDEF'):
+        for dxftype in ('TEXT', 'ATTRIB', 'ATTDEF'):
             dispatch_table[dxftype] = self.draw_text_entity
         for dxftype in ('CIRCLE', 'ARC', 'ELLIPSE', 'SPLINE'):
             dispatch_table[dxftype] = self.draw_curve_entity
@@ -103,8 +107,8 @@ class Frontend:
         # These types have a virtual_entities() method, which returns
         # the content of the associated block or anonymous block
         for dxftype in ['INSERT', 'DIMENSION', 'ARC_DIMENSION',
-                        'LARGE_RADIAL_DIMENSION', 'LEADER',
-                        'MLINE', 'ACAD_TABLE']:
+            'LARGE_RADIAL_DIMENSION', 'LEADER',
+            'MLINE', 'ACAD_TABLE']:
             dispatch_table[dxftype] = self.draw_composite_entity
 
         return dispatch_table
@@ -116,7 +120,7 @@ class Frontend:
         self.log_message(f'skipped entity {str(entity)}. Reason: "{msg}"')
 
     def override_properties(self, entity: DXFGraphic,
-                            properties: Properties) -> None:
+        properties: Properties) -> None:
         """ The :meth:`override_properties` filter can change the properties of
         an entity independent from the DXF attributes.
 
@@ -125,35 +129,48 @@ class Frontend:
         `properties` object. It is recommended to modify only the `properties`
         object in this filter.
         """
-        if entity.dxftype() == 'HATCH':
-            properties.color = set_color_alpha(properties.color, 200)
 
-    def draw_layout(self, layout: 'Layout', finalize: bool = True) -> None:
+    def draw_layout(
+        self, layout: 'Layout',
+        finalize: bool = True,
+        *,
+        filter_func: FilterFunc = None
+    ) -> None:
         self.parent_stack = []
         handle_mapping = list(layout.get_redraw_order())
         if handle_mapping:
-            self.draw_entities(reorder.ascending(layout, handle_mapping))
+            self.draw_entities(
+                reorder.ascending(layout, handle_mapping),
+                filter_func=filter_func,
+            )
         else:
-            self.draw_entities(layout)
+            self.draw_entities(
+                layout,
+                filter_func=filter_func,
+            )
         self.out.set_background(self.ctx.current_layout.background_color)
         if finalize:
             self.out.finalize()
 
-    def draw_entities(self, entities: Iterable[DXFGraphic]) -> None:
+    def draw_entities(
+        self,
+        entities: Iterable[DXFGraphic],
+        *,
+        filter_func: FilterFunc = None,
+    ) -> None:
+        if filter_func is not None:
+            entities = filter(filter_func, entities)
         for entity in entities:
             # Skip unsupported DXF entities - just tag storage to preserve data
             if isinstance(entity, DXFTagStorage):
                 self.skip_entity(entity, 'Cannot parse DXF entity')
                 continue
-
             properties = self.ctx.resolve_all(entity)
             self.override_properties(entity, properties)
 
-            # The content of a block reference does not depend
-            # on the visibility state of the INSERT entity:
-            if properties.is_visible or entity.dxftype() == 'INSERT':
+            if properties.is_visible:
                 self.draw_entity(entity, properties)
-            elif not properties.is_visible:
+            else:
                 self.skip_entity(entity, 'invisible')
 
     def draw_entity(self, entity: DXFGraphic, properties: Properties) -> None:
@@ -179,7 +196,7 @@ class Frontend:
         self.out.exit_entity(entity)
 
     def draw_line_entity(self, entity: DXFGraphic,
-                         properties: Properties) -> None:
+        properties: Properties) -> None:
         d, dxftype = entity.dxf, entity.dxftype()
         if dxftype == 'LINE':
             self.out.draw_line(d.start, d.end, properties)
@@ -189,36 +206,74 @@ class Frontend:
             delta = d.unit_vector * INFINITE_LINE_LENGTH
             if dxftype == 'XLINE':
                 self.out.draw_line(start - delta / 2, start + delta / 2,
-                                   properties)
+                    properties)
             elif dxftype == 'RAY':
                 self.out.draw_line(start, start + delta, properties)
         else:
             raise TypeError(dxftype)
 
     def draw_text_entity(self, entity: DXFGraphic,
-                         properties: Properties) -> None:
+        properties: Properties) -> None:
         if is_spatial_text(Vec3(entity.dxf.extrusion)):
             self.draw_text_entity_3d(entity, properties)
         else:
             self.draw_text_entity_2d(entity, properties)
 
     def draw_text_entity_2d(self, entity: DXFGraphic,
-                            properties: Properties) -> None:
+        properties: Properties) -> None:
         d, dxftype = entity.dxf, entity.dxftype()
-        if dxftype in ('TEXT', 'MTEXT', 'ATTRIB', 'ATTDEF'):
-            entity = cast(Union[Text, MText, Attrib, AttDef], entity)
+        if dxftype in ('TEXT', 'ATTRIB', 'ATTDEF'):
+            entity = cast(Union[Text, Attrib, AttDef], entity)
             for line, transform, cap_height in simplified_text_chunks(
-                    entity, self.out, font=properties.font):
+                entity, self.out, font=properties.font):
                 self.out.draw_text(line, transform, properties, cap_height)
         else:
             raise TypeError(dxftype)
 
     def draw_text_entity_3d(self, entity: DXFGraphic,
-                            properties: Properties) -> None:
+        properties: Properties) -> None:
         self.skip_entity(entity, '3D text not supported')
 
+    def draw_mtext_entity(self, mtext: 'MText',
+        properties: Properties) -> None:
+        if is_spatial_text(Vec3(mtext.dxf.extrusion)):
+            self.skip_entity(mtext, '3D MTEXT not supported')
+            return
+        if mtext.has_columns:
+            columns = mtext.columns
+            if len(columns.linked_columns):
+                has_linked_content = any(c.text for c in columns.linked_columns)
+                if has_linked_content:
+                    # Column content is spread across multiple MTEXT entities.
+                    # For now we trust the DXF creator that each MTEXT entity
+                    # has exact the required column content.
+                    # This is not granted and AutoCAD/BricsCAD do the column
+                    # content distribution always by themself!
+                    self.draw_mtext_column(mtext, properties)
+                    for column in mtext.columns.linked_columns:
+                        self.draw_mtext_column(column, properties)
+                    return
+            self.distribute_mtext_columns_content(mtext, properties)
+        else:
+            self.draw_mtext_column(mtext, properties)
+
+    def distribute_mtext_columns_content(self, mtext: MText,
+        properties: Properties):
+        """ Distribute the content of the MTEXT entity across multiple columns
+        """
+        # TODO: complex MTEXT renderer
+        self.draw_mtext_column(mtext, properties)
+
+    def draw_mtext_column(self, mtext: MText,
+        properties: Properties) -> None:
+        """ Draw the content of a MTEXT entity as a single column. """
+        # TODO: complex MTEXT renderer
+        for line, transform, cap_height in simplified_text_chunks(
+            mtext, self.out, font=properties.font):
+            self.out.draw_text(line, transform, properties, cap_height)
+
     def draw_curve_entity(self, entity: DXFGraphic,
-                          properties: Properties) -> None:
+        properties: Properties) -> None:
         try:
             path = make_path(entity)
         except AttributeError:  # API usage error
@@ -227,7 +282,7 @@ class Frontend:
         self.out.draw_path(path, properties)
 
     def draw_point_entity(self, entity: DXFGraphic,
-                          properties: Properties) -> None:
+        properties: Properties) -> None:
         point = cast('Point', entity)
         pdmode = self.out.pdmode
 
@@ -255,32 +310,31 @@ class Frontend:
                         self.out.draw_line(start, end, properties)
                     pass
                 else:  # CIRCLE
-                    self.draw_elliptic_arc_entity(entity, properties)
+                    self.draw_curve_entity(entity, properties)
 
     def draw_solid_entity(self, entity: DXFGraphic,
-                          properties: Properties) -> None:
-        # Handles SOLID, TRACE and 3DFACE
+        properties: Properties) -> None:
+        assert isinstance(entity, (Solid, Face3d)), \
+            "API error, requires a SOLID, TRACE or 3DFACE entity"
         dxf, dxftype = entity.dxf, entity.dxftype()
-        points = get_tri_or_quad_points(
-            entity, adjust_order=dxftype != '3DFACE')
-        # TRACE is an OCS entity
-        if dxftype == 'TRACE' and dxf.hasattr('extrusion'):
-            ocs = entity.ocs()
-            points = list(ocs.points_to_wcs(points))
+        points = entity.wcs_vertices()
         if dxftype == '3DFACE':
-            self.out.draw_path(Path.from_vertices(points, close=True),
-                               properties)
+            self.out.draw_path(from_vertices(points, close=True), properties)
         else:
-            # Set default SOLID filling for SOLID and TRACE
+            # set solid fill type for SOLID and TRACE
             properties.filling = Filling()
             self.out.draw_filled_polygon(points, properties)
 
     def draw_hatch_entity(self, entity: DXFGraphic,
-                          properties: Properties) -> None:
-        def to_path(p):
-            path = Path.from_hatch_boundary_path(p, ocs, elevation)
-            path.close()
-            return path
+        properties: Properties) -> None:
+        def to_path(paths):
+            loops = []
+            for boundary in paths:
+                path = from_hatch_boundary_path(boundary, ocs, elevation)
+                for sub_path in path.sub_paths():
+                    sub_path.close()
+                    loops.append(sub_path)
+            return loops
 
         if not self.out.show_hatch:
             return
@@ -295,14 +349,14 @@ class Frontend:
         holes = []
         paths = hatch.paths.rendering_paths(hatch.dxf.hatch_style)
         if self.nested_polygon_detection:
-            polygons = self.nested_polygon_detection(map(to_path, paths))
-            external_paths, holes = nesting.winding_deconstruction(polygons)
+            polygons = self.nested_polygon_detection(to_path(paths))
+            external_paths, holes = winding_deconstruction(polygons)
         else:
             for p in paths:
                 if p.path_type_flags & const.BOUNDARY_PATH_EXTERNAL:
-                    external_paths.append(to_path(p))
+                    external_paths.extend(to_path(p))
                 else:
-                    holes.append(to_path(p))
+                    holes.extend(to_path(p))
 
         if external_paths:
             self.out.draw_filled_paths(external_paths, holes, properties)
@@ -311,7 +365,7 @@ class Frontend:
             self.out.draw_filled_paths([holes[0]], holes[1:], properties)
 
     def draw_wipeout_entity(self, entity: DXFGraphic,
-                            properties: Properties) -> None:
+        properties: Properties) -> None:
         wipeout = cast(Wipeout, entity)
         properties.filling = Filling()
         properties.color = self.ctx.current_layout.background_color
@@ -319,7 +373,7 @@ class Frontend:
         self.out.draw_filled_polygon(path, properties)
 
     def draw_viewport_entity(self, entity: DXFGraphic,
-                             properties: Properties) -> None:
+        properties: Properties) -> None:
         assert entity.dxftype() == 'VIEWPORT'
         # Special VIEWPORT id == 1, this viewport defines the "active viewport"
         # which is the area currently shown in the layout tab by the CAD
@@ -355,21 +409,21 @@ class Frontend:
         # Set default SOLID filling for VIEWPORT
         props.filling = Filling()
         self.out.draw_filled_polygon([Vec3(x, y, 0) for x, y in points],
-                                     props)
+            props)
 
     def draw_mesh_entity(self, entity: DXFGraphic,
-                         properties: Properties) -> None:
+        properties: Properties) -> None:
         builder = MeshBuilder.from_mesh(entity)
         self.draw_mesh_builder_entity(builder, properties)
 
     def draw_mesh_builder_entity(self, builder: MeshBuilder,
-                                 properties: Properties) -> None:
+        properties: Properties) -> None:
         for face in builder.faces_as_vertices():
             self.out.draw_path(
-                Path.from_vertices(face, close=True), properties=properties)
+                from_vertices(face, close=True), properties=properties)
 
     def draw_polyline_entity(self, entity: DXFGraphic,
-                             properties: Properties) -> None:
+        properties: Properties) -> None:
         dxftype = entity.dxftype()
         if dxftype == 'POLYLINE':
             e = cast(Polyface, entity)
@@ -413,7 +467,7 @@ class Frontend:
         self.out.draw_path(path, properties)
 
     def draw_composite_entity(self, entity: DXFGraphic,
-                              properties: Properties) -> None:
+        properties: Properties) -> None:
         def set_opaque(entities: Iterable[DXFGraphic]):
             for child in entities:
                 # todo: defaults to 1.0 (fully transparent)???
