@@ -32,10 +32,12 @@ from .clipping import *
 
 Vertex = Union[Sequence[float], Vec3, Vec2]
 VecXY = Union[Vec2, Vec3]  # Vector with x and y attributes
+ABS_TOL = 1e-12
+REL_TOL = 1e-9
 
 
 def close_vectors(a: Iterable[VecXY], b: Iterable[Vertex], *,
-                  rel_tol=1e-9, abs_tol=1e-12) -> bool:
+                  rel_tol=REL_TOL, abs_tol=ABS_TOL) -> bool:
     return all(v1.isclose(v2, rel_tol=rel_tol, abs_tol=abs_tol)
                for v1, v2 in zip(a, b))
 
