@@ -1,15 +1,30 @@
 # Copyright (c) 2020-2021, Manfred Moitzi
 # License: MIT License
 from typing import (
-    TYPE_CHECKING, List, Iterable, Tuple, Optional, Dict, Sequence
+    TYPE_CHECKING,
+    List,
+    Iterable,
+    Tuple,
+    Optional,
+    Dict,
+    Sequence,
 )
 
 import math
 import itertools
 from ezdxf.math import (
-    Vec3, Z_AXIS, OCS, Matrix44, BoundingBox, ConstructionEllipse,
-    cubic_bezier_from_ellipse, Bezier4P, Bezier3P, BSpline,
-    reverse_bezier_curves, bulge_to_arc
+    Vec3,
+    Z_AXIS,
+    OCS,
+    Matrix44,
+    BoundingBox,
+    ConstructionEllipse,
+    cubic_bezier_from_ellipse,
+    Bezier4P,
+    Bezier3P,
+    BSpline,
+    reverse_bezier_curves,
+    bulge_to_arc,
 )
 
 from ezdxf.query import EntityQuery
@@ -22,20 +37,59 @@ if TYPE_CHECKING:
     from ezdxf.eztypes import Vertex, Layout, EntityQuery
 
 __all__ = [
-    'bbox', 'fit_paths_into_box', 'transform_paths', 'transform_paths_to_ocs',
-    'render_lwpolylines', 'render_polylines2d', 'render_polylines3d',
-    'render_lines', 'render_hatches', 'render_splines_and_polylines',
-    'add_bezier4p', 'add_bezier3p', 'add_ellipse', 'add_2d_polyline',
-    'add_spline'
+    "bbox",
+    "fit_paths_into_box",
+    "transform_paths",
+    "transform_paths_to_ocs",
+    "render_lwpolylines",
+    "render_polylines2d",
+    "render_polylines3d",
+    "render_lines",
+    "render_hatches",
+    "render_mpolygons",
+    "render_splines_and_polylines",
+    "add_bezier4p",
+    "add_bezier3p",
+    "add_ellipse",
+    "add_2d_polyline",
+    "add_spline",
+    "to_multi_path",
+    "single_paths",
 ]
 
 MAX_DISTANCE = 0.01
 MIN_SEGMENTS = 4
 G1_TOL = 1e-4
+IS_CLOSE_TOL = 1e-10
+
+
+def to_multi_path(paths: Iterable[Path]) -> Path:
+    """Returns a multi-path object from all given paths and their sub-paths.
+
+    .. versionadded:: 0.17
+
+    """
+    multi_path = Path()
+    for p in paths:
+        multi_path.extend_multi_path(p)
+    return multi_path
+
+
+def single_paths(paths: Iterable[Path]) -> Iterable[Path]:
+    """Yields all given paths and their sub-paths as single path objects.
+
+    .. versionadded:: 0.17
+
+    """
+    for p in paths:
+        if p.has_sub_paths:
+            yield from p.sub_paths()
+        else:
+            yield p
 
 
 def transform_paths(paths: Iterable[Path], m: Matrix44) -> List[Path]:
-    """ Transform multiple :class:`Path` objects at once by transformation
+    """Transform multiple :class:`Path` objects at once by transformation
     matrix `m`. Returns a list of the transformed :class:`Path` objects.
 
     Args:
@@ -52,25 +106,31 @@ def transform_paths(paths: Iterable[Path], m: Matrix44) -> List[Path]:
             vertices.extend(cmd)
 
     def rebuild(vertices):
+        # localize variables:
+        start_path, line_to, curve3_to, curve4_to, move_to = Command
+
         path = None
         collect = []
         for vertex, cmd in zip(vertices, commands):
-            if cmd == Command.START_PATH:
+            if cmd == start_path:
                 if path is not None:
                     transformed_paths.append(path)
                 path = Path(vertex)
-            elif cmd == Command.LINE_TO:
+            elif cmd == line_to:
                 path.line_to(vertex)
-            elif cmd == Command.CURVE3_TO:
+            elif cmd == curve3_to:
                 collect.append(vertex)
                 if len(collect) == 2:
                     path.curve3_to(collect[0], collect[1])
                     collect.clear()
-            elif cmd == Command.CURVE4_TO:
+            elif cmd == curve4_to:
                 collect.append(vertex)
                 if len(collect) == 3:
                     path.curve4_to(collect[0], collect[1], collect[2])
                     collect.clear()
+            elif cmd == move_to:
+                path.move_to(vertex)
+
         if path is not None:
             transformed_paths.append(path)
 
@@ -86,7 +146,7 @@ def transform_paths(paths: Iterable[Path], m: Matrix44) -> List[Path]:
 
 
 def transform_paths_to_ocs(paths: Iterable[Path], ocs: OCS) -> List[Path]:
-    """ Transform multiple :class:`Path` objects at once from WCS to OCS.
+    """Transform multiple :class:`Path` objects at once from WCS to OCS.
     Returns a list of the transformed :class:`Path` objects.
 
     Args:
@@ -99,15 +159,16 @@ def transform_paths_to_ocs(paths: Iterable[Path], ocs: OCS) -> List[Path]:
     return transform_paths(paths, t)
 
 
-def bbox(paths: Iterable[Path], flatten=0.01,
-         segments: int = 16) -> BoundingBox:
-    """ Returns the :class:`~ezdxf.math.BoundingBox` for the given paths.
+def bbox(
+    paths: Iterable[Path], flatten=0.01, segments: int = 16
+) -> BoundingBox:
+    """Returns the :class:`~ezdxf.math.BoundingBox` for the given paths.
 
     Args:
         paths: iterable of :class:`~ezdxf.path.Path` objects
         flatten: value != 0  for bounding box calculation from the flattened
-        path and value == 0 for bounding box from the control vertices.
-        Default value is 0.01 as max flattening distance.
+            path and value == 0 for bounding box from the control vertices.
+            Default value is 0.01 as max flattening distance.
         segments: minimal segment count for flattening
 
     """
@@ -120,11 +181,13 @@ def bbox(paths: Iterable[Path], flatten=0.01,
     return box
 
 
-def fit_paths_into_box(paths: Iterable[Path],
-                       size: Tuple[float, float, float],
-                       uniform: bool = True,
-                       source_box: BoundingBox = None) -> List[Path]:
-    """ Scale the given `paths` to fit into a box of the given `size`,
+def fit_paths_into_box(
+    paths: Iterable[Path],
+    size: Tuple[float, float, float],
+    uniform: bool = True,
+    source_box: BoundingBox = None,
+) -> List[Path]:
+    """Scale the given `paths` to fit into a box of the given `size`,
     so that all path vertices are inside this borders.
     If `source_box` is ``None`` the default source bounding box is calculated
     from the control points of the `paths`.
@@ -152,7 +215,7 @@ def fit_paths_into_box(paths: Iterable[Path],
         return paths
     target_size = Vec3(size)
     if target_size == (0, 0, 0) or min(target_size) < 0:
-        raise ValueError('invalid target size')
+        raise ValueError("invalid target size")
 
     if uniform:
         sx, sy, sz = _get_uniform_scaling(current_box.size, target_size)
@@ -176,7 +239,7 @@ def _get_uniform_scaling(current_size: Vec3, target_size: Vec3):
 
     uniform_scale = min(scale_x, scale_y, scale_z)
     if uniform_scale is math.inf:
-        raise ArithmeticError('internal error')
+        raise ArithmeticError("internal error")
     scale_x = uniform_scale if target_size.x > TOL else 0
     scale_y = uniform_scale if target_size.y > TOL else 0
     scale_z = uniform_scale if target_size.z > TOL else 0
@@ -201,14 +264,15 @@ def _get_non_uniform_scaling(current_size: Vec3, target_size: Vec3):
 
 
 def render_lwpolylines(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        distance: float = MAX_DISTANCE,
-        segments: int = MIN_SEGMENTS,
-        extrusion: 'Vertex' = Z_AXIS,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    distance: float = MAX_DISTANCE,
+    segments: int = MIN_SEGMENTS,
+    extrusion: "Vertex" = Z_AXIS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as
     :class:`~ezdxf.entities.LWPolyline` entities.
     The `extrusion` vector is applied to all paths, all vertices are projected
     onto the plane normal to this extrusion vector. The default extrusion vector
@@ -229,27 +293,30 @@ def render_lwpolylines(
     .. versionadded:: 0.16
 
     """
-    lwpolylines = list(converter.to_lwpolylines(
-        paths,
-        distance=distance,
-        segments=segments,
-        extrusion=extrusion,
-        dxfattribs=dxfattribs,
-    ))
+    lwpolylines = list(
+        converter.to_lwpolylines(
+            paths,
+            distance=distance,
+            segments=segments,
+            extrusion=extrusion,
+            dxfattribs=dxfattribs,
+        )
+    )
     for lwpolyline in lwpolylines:
         layout.add_entity(lwpolyline)
     return EntityQuery(lwpolylines)
 
 
 def render_polylines2d(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        distance: float = 0.01,
-        segments: int = 4,
-        extrusion: 'Vertex' = Z_AXIS,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as 2D
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    distance: float = 0.01,
+    segments: int = 4,
+    extrusion: "Vertex" = Z_AXIS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as 2D
     :class:`~ezdxf.entities.Polyline` entities.
     The `extrusion` vector is applied to all paths, all vertices are projected
     onto the plane normal to this extrusion vector.The default extrusion vector
@@ -270,29 +337,32 @@ def render_polylines2d(
     .. versionadded:: 0.16
 
     """
-    polylines2d = list(converter.to_polylines2d(
-        paths,
-        distance=distance,
-        segments=segments,
-        extrusion=extrusion,
-        dxfattribs=dxfattribs,
-    ))
+    polylines2d = list(
+        converter.to_polylines2d(
+            paths,
+            distance=distance,
+            segments=segments,
+            extrusion=extrusion,
+            dxfattribs=dxfattribs,
+        )
+    )
     for polyline2d in polylines2d:
         layout.add_entity(polyline2d)
     return EntityQuery(polylines2d)
 
 
 def render_hatches(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        edge_path: bool = True,
-        distance: float = MAX_DISTANCE,
-        segments: int = MIN_SEGMENTS,
-        g1_tol: float = G1_TOL,
-        extrusion: 'Vertex' = Z_AXIS,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    edge_path: bool = True,
+    distance: float = MAX_DISTANCE,
+    segments: int = MIN_SEGMENTS,
+    g1_tol: float = G1_TOL,
+    extrusion: "Vertex" = Z_AXIS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as
     :class:`~ezdxf.entities.Hatch` entities.
     The `extrusion` vector is applied to all paths, all vertices are projected
     onto the plane normal to this extrusion vector. The default extrusion vector
@@ -305,7 +375,7 @@ def render_hatches(
         edge_path: ``True`` for edge paths build of LINE and SPLINE edges,
             ``False`` for only LWPOLYLINE paths as boundary paths
         distance:  maximum distance, see :meth:`Path.flattening`
-        segments: minimum segment count per Bézier curve to flatten LWPOLYLINE paths
+        segments: minimum segment count per Bézier curve to flatten polyline paths
         g1_tol: tolerance for G1 continuity check to separate SPLINE edges
         extrusion: extrusion vector for all paths
         dxfattribs: additional DXF attribs
@@ -316,28 +386,77 @@ def render_hatches(
     .. versionadded:: 0.16
 
     """
-    hatches = list(converter.to_hatches(
-        paths,
-        edge_path=edge_path,
-        distance=distance,
-        segments=segments,
-        g1_tol=g1_tol,
-        extrusion=extrusion,
-        dxfattribs=dxfattribs,
-    ))
+    hatches = list(
+        converter.to_hatches(
+            paths,
+            edge_path=edge_path,
+            distance=distance,
+            segments=segments,
+            g1_tol=g1_tol,
+            extrusion=extrusion,
+            dxfattribs=dxfattribs,
+        )
+    )
     for hatch in hatches:
         layout.add_entity(hatch)
     return EntityQuery(hatches)
 
 
+def render_mpolygons(
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    distance: float = MAX_DISTANCE,
+    segments: int = MIN_SEGMENTS,
+    extrusion: "Vertex" = Z_AXIS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as
+    :class:`~ezdxf.entities.MPolygon` entities. The MPOLYGON entity supports
+    only polyline boundary paths. All curves will be approximated.
+
+    The `extrusion` vector is applied to all paths, all vertices are projected
+    onto the plane normal to this extrusion vector. The default extrusion vector
+    is the WCS z-axis. The plane elevation is the distance from the WCS origin
+    to the start point of the first path.
+
+    Args:
+        layout: the modelspace, a paperspace layout or a block definition
+        paths: iterable of :class:`Path` objects
+        distance:  maximum distance, see :meth:`Path.flattening`
+        segments: minimum segment count per Bézier curve to flatten polyline paths
+        extrusion: extrusion vector for all paths
+        dxfattribs: additional DXF attribs
+
+    Returns:
+        created entities in an :class:`~ezdxf.query.EntityQuery` object
+
+    .. versionadded:: 0.17
+
+    """
+    polygons = list(
+        converter.to_mpolygons(
+            paths,
+            distance=distance,
+            segments=segments,
+            extrusion=extrusion,
+            dxfattribs=dxfattribs,
+        )
+    )
+    for polygon in polygons:
+        layout.add_entity(polygon)
+    return EntityQuery(polygons)
+
+
 def render_polylines3d(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        distance: float = MAX_DISTANCE,
-        segments: int = MIN_SEGMENTS,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as 3D
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    distance: float = MAX_DISTANCE,
+    segments: int = MIN_SEGMENTS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as 3D
     :class:`~ezdxf.entities.Polyline` entities.
 
     Args:
@@ -354,25 +473,28 @@ def render_polylines3d(
 
     """
 
-    polylines3d = list(converter.to_polylines3d(
-        paths,
-        distance=distance,
-        segments=segments,
-        dxfattribs=dxfattribs,
-    ))
+    polylines3d = list(
+        converter.to_polylines3d(
+            paths,
+            distance=distance,
+            segments=segments,
+            dxfattribs=dxfattribs,
+        )
+    )
     for polyline3d in polylines3d:
         layout.add_entity(polyline3d)
     return EntityQuery(polylines3d)
 
 
 def render_lines(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        distance: float = MAX_DISTANCE,
-        segments: int = MIN_SEGMENTS,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    distance: float = MAX_DISTANCE,
+    segments: int = MIN_SEGMENTS,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as
     :class:`~ezdxf.entities.Line` entities.
 
     Args:
@@ -388,24 +510,27 @@ def render_lines(
     .. versionadded:: 0.16
 
     """
-    lines = list(converter.to_lines(
-        paths,
-        distance=distance,
-        segments=segments,
-        dxfattribs=dxfattribs,
-    ))
+    lines = list(
+        converter.to_lines(
+            paths,
+            distance=distance,
+            segments=segments,
+            dxfattribs=dxfattribs,
+        )
+    )
     for line in lines:
         layout.add_entity(line)
     return EntityQuery(lines)
 
 
 def render_splines_and_polylines(
-        layout: 'Layout',
-        paths: Iterable[Path],
-        *,
-        g1_tol: float = G1_TOL,
-        dxfattribs: Optional[Dict] = None) -> EntityQuery:
-    """ Render the given `paths` into `layout` as :class:`~ezdxf.entities.Spline`
+    layout: "Layout",
+    paths: Iterable[Path],
+    *,
+    g1_tol: float = G1_TOL,
+    dxfattribs: Optional[Dict] = None
+) -> EntityQuery:
+    """Render the given `paths` into `layout` as :class:`~ezdxf.entities.Spline`
     and 3D :class:`~ezdxf.entities.Polyline` entities.
 
     Args:
@@ -420,19 +545,22 @@ def render_splines_and_polylines(
     .. versionadded:: 0.16
 
     """
-    entities = list(converter.to_splines_and_polylines(
-        paths,
-        g1_tol=g1_tol,
-        dxfattribs=dxfattribs,
-    ))
+    entities = list(
+        converter.to_splines_and_polylines(
+            paths,
+            g1_tol=g1_tol,
+            dxfattribs=dxfattribs,
+        )
+    )
     for entity in entities:
         layout.add_entity(entity)
     return EntityQuery(entities)
 
 
-def add_ellipse(path: Path, ellipse: ConstructionEllipse, segments=1,
-                reset=True) -> None:
-    """ Add an elliptical arc as multiple cubic Bèzier-curves to the given
+def add_ellipse(
+    path: Path, ellipse: ConstructionEllipse, segments=1, reset=True
+) -> None:
+    """Add an elliptical arc as multiple cubic Bèzier-curves to the given
     `path`, use :meth:`~ezdxf.math.ConstructionEllipse.from_arc` constructor
     of class :class:`~ezdxf.math.ConstructionEllipse` to add circular arcs.
 
@@ -462,7 +590,7 @@ def add_ellipse(path: Path, ellipse: ConstructionEllipse, segments=1,
 
 
 def add_bezier4p(path: Path, curves: Iterable[Bezier4P]) -> None:
-    """ Add multiple cubic Bèzier-curves to the given `path`.
+    """Add multiple cubic Bèzier-curves to the given `path`.
 
     Auto-detect the connection point to the given `path`, if neither the start-
     nor the end point of the curves is close to the path end point, a line from
@@ -474,6 +602,8 @@ def add_bezier4p(path: Path, curves: Iterable[Bezier4P]) -> None:
         add linear Bézier curve segments as LINE_TO commands
 
     """
+    rel_tol = 1e-15
+    abs_tol = 0.0
     curves = list(curves)
     if not len(curves):
         return
@@ -484,19 +614,21 @@ def add_bezier4p(path: Path, curves: Iterable[Bezier4P]) -> None:
 
     for curve in curves:
         start, ctrl1, ctrl2, end = curve.control_points
-        if not start.isclose(path.end, abs_tol=1e-9):
+        if not start.isclose(path.end):
             path.line_to(start)
 
         # add linear bezier segments as LINE_TO commands
-        if start.isclose(ctrl1, abs_tol=1e-9) and \
-                end.isclose(ctrl2, abs_tol=1e-9):
+        if (
+            start.isclose(ctrl1, rel_tol=rel_tol, abs_tol=abs_tol)
+            and end.isclose(ctrl2, rel_tol=rel_tol, abs_tol=abs_tol)
+        ):
             path.line_to(end)
         else:
             path.curve4_to(end, ctrl1, ctrl2)
 
 
 def add_bezier3p(path: Path, curves: Iterable[Bezier3P]) -> None:
-    """ Add multiple quadratic Bèzier-curves to the given `path`.
+    """Add multiple quadratic Bèzier-curves to the given `path`.
 
     Auto-detect the connection point to the given `path`, if neither the start-
     nor the end point of the curves is close to the path end point, a line from
@@ -508,6 +640,8 @@ def add_bezier3p(path: Path, curves: Iterable[Bezier3P]) -> None:
         add linear Bézier curve segments as LINE_TO commands
 
     """
+    rel_tol = 1e-15
+    abs_tol = 0.0
     curves = list(curves)
     if not len(curves):
         return
@@ -518,41 +652,50 @@ def add_bezier3p(path: Path, curves: Iterable[Bezier3P]) -> None:
 
     for curve in curves:
         start, ctrl, end = curve.control_points
-        if not start.isclose(path.end, abs_tol=1e-9):
+        if not start.isclose(path.end, rel_tol=rel_tol, abs_tol=abs_tol):
             path.line_to(start)
 
-        # add linear bezier segments as LINE_TO commands
-        if start.isclose(ctrl, abs_tol=1e-9) or end.isclose(ctrl, abs_tol=1e-9):
+        if (
+            start.isclose(ctrl, rel_tol=rel_tol, abs_tol=abs_tol)
+            or end.isclose(ctrl, rel_tol=rel_tol, abs_tol=abs_tol)
+        ):
             path.line_to(end)
         else:
             path.curve3_to(end, ctrl)
 
 
-def add_2d_polyline(path: Path, points: Iterable[Sequence[float]], close: bool,
-                    ocs: OCS, elevation: float) -> None:
-    """ Internal API to add 2D polylines which may include bulges to an
+def add_2d_polyline(
+    path: Path,
+    points: Iterable[Sequence[float]],
+    close: bool,
+    ocs: OCS,
+    elevation: float,
+) -> None:
+    """Internal API to add 2D polylines which may include bulges to an
     **empty** path.
 
     """
 
     def bulge_to(p1: Vec3, p2: Vec3, bulge: float):
-        if p1.isclose(p2):
+        if p1.isclose(p2, rel_tol=IS_CLOSE_TOL, abs_tol=0):
             return
         center, start_angle, end_angle, radius = bulge_to_arc(p1, p2, bulge)
         ellipse = ConstructionEllipse.from_arc(
-            center, radius, Z_AXIS,
+            center,
+            radius,
+            Z_AXIS,
             math.degrees(start_angle),
             math.degrees(end_angle),
         )
         curves = list(cubic_bezier_from_ellipse(ellipse))
         curve0 = curves[0]
         cp0 = curve0.control_points[0]
-        if cp0.isclose(p2):
+        if cp0.isclose(p2, rel_tol=IS_CLOSE_TOL, abs_tol=0):
             curves = reverse_bezier_curves(curves)
         add_bezier4p(path, curves)
 
     if len(path):
-        raise ValueError('Requires an empty path.')
+        raise ValueError("Requires an empty path.")
 
     prev_point = None
     prev_bulge = 0
@@ -574,7 +717,9 @@ def add_2d_polyline(path: Path, points: Iterable[Sequence[float]], close: bool,
         prev_point = point
         prev_bulge = bulge
 
-    if close and not path.start.isclose(path.end):
+    if close and not path.start.isclose(
+        path.end, rel_tol=IS_CLOSE_TOL, abs_tol=0
+    ):
         if prev_bulge:
             bulge_to(path.end, path.start, prev_bulge)
         else:
@@ -585,7 +730,7 @@ def add_2d_polyline(path: Path, points: Iterable[Sequence[float]], close: bool,
 
 
 def add_spline(path: Path, spline: BSpline, level=4, reset=True) -> None:
-    """ Add a B-spline as multiple cubic Bèzier-curves.
+    """Add a B-spline as multiple cubic Bèzier-curves.
 
     Non-rational B-splines of 3rd degree gets a perfect conversion to
     cubic bezier curves with a minimal count of curve segments, all other
@@ -610,8 +755,7 @@ def add_spline(path: Path, spline: BSpline, level=4, reset=True) -> None:
     if len(path) == 0 and reset:
         path.start = spline.point(0)
     if spline.degree == 3 and not spline.is_rational and spline.is_clamped:
-        curves = [Bezier4P(points) for points in
-                  spline.bezier_decomposition()]
+        curves = [Bezier4P(points) for points in spline.bezier_decomposition()]
     else:
         curves = spline.cubic_bezier_approximation(level=level)
     add_bezier4p(path, curves)
